
#############################################################################
#
# Computes all extensions of the elementary abelian module under the action
# of <pc> described by <m>
#
# Example input: pc := SmallGroup(18,4); p := 3;
#                m := One(GF(p))*[[[1,2],[0,2]],[[1,1],[0,1]],[[1,0],[0,1]]];
#

InstallGlobalFunction(DerivedSeriesExtensionsSQ,
function(pc, p, m)
    local ser,       # derived series of <pc>
          der,       # indices of generators of last non-trivial term of <ser>
          pairs,     # index pairs corresponding to entries in <der>
          d,         # dimension of module
          M,         # module defined by <m>
          cc,        # 2-cohomology record
          basis,     # vector space basis of module A = (F_p)^d
          vec,       # a vector in <basis>
          comm,      # list of generating commutators of [A, N]
          ind,       # an index from <der>
          subbasis,  # basis of subspace [A, N] of the module A
          D,         # Aut(G) x Aut(A)
          tup,       # pseudo random element in <D>
          C,         # group of compatible pairs
          mats,      # generators of <C> as matrix group
          Mgrp,      # <C> as matrix group
          orbs,      # orbits under <Mgrp>
          ccvecs,    # list of cocycles, representatives of <orbs>
          result,    # list of parts of cocycles corresponding to N x N
          rels,      # list of index pairs of relators of <pc>
          inds,      # list of decisive positions of index pairs in <rels>
          parts,     # decisive parts of generators of group of cocycles
#          pair,      # one index pair
#          pos,       # position of <pair> in <rels> 
          i,         # loop counter
          c,         # cocycle from <ccvecs>
          positions, # list of indices of cocycles in <ccvecs>
          vs;        # subspace of A generated by [A, N] and <i>-th cocycle

    ### MAIN FUNCTION BODY ################################################

    if not IsSolvable(pc) then
        Error("<pc> has to be a solvable group");
    fi;

    ## Get the derived series of the input group. We assume that the generating
    ## set runs through the series. The indices <der> of the generators of the
    ## last non-trivial term (denote by N) are needed later.  
    ser := DerivedSeries(pc);
    ## test whether the series runs through the last non-trivial term N of 
    ## the derived series and adjust if necessary
    if not ForAll(Pcgs(ser[Length(ser)-1]), elm-> elm in Pcgs(pc)) then
        ser := DerivedSeries(PcGroupToPcpGroup(pc));
        pc := PcpGroupToPcGroup(PcpGroupBySeries(ser));
        ser := DerivedSeries(pc);        
    fi;
    # indices of generators of N
    der := List(Pcgs(ser[Length(ser)-1]), x-> Position(Pcgs(pc), x));
    pairs := List(Combinations(der, 2), Reversed);

    ## compute the (outer) commutator of N and the module A as subspace of A
    d := Size(m[1]);
    basis := IdentityMat(d, GF(p));
    comm := EmptyPlist(Size(basis)*Length(der));
    for vec in basis do
        for ind in der do
            # a^n - a for a in A, n in N
            Add(comm, vec^m[ind]-vec);
        od;
    od;
    # basis of subspace
    vs := (VectorSpace(GF(p), comm));
    if Dimension(vs) = 0 then
        Info(InfoSpecialExt, 1, "[A,N] = 0");
    fi;
    subbasis := Basis(VectorSpace(GF(p), comm));

    ## compare needed dimension from cocycle parts with maximal possible one 
    if d > Length(subbasis) + Length(pairs) then
        # image of part of cocycle never equals A/[A,N]
        Info(InfoSpecialExt, 1, "d > dim([A,N]) + #gens(N)^2 - #gens(N)");
        return EmptyPlist(0);
    fi;

    ## construct module
    M := GModuleByMats(m, GF(p));

    ## test if A = [N, A], then only the split extension exists
    ## (Lemma 3.1, Eick, Wright, Computing subgroups by exhibition in finite
    ##  solvable groups, J. Symb. Comp. (2006), 11)
    if Size(subbasis) = d then
        Info(InfoSpecialExt, 1, "[A,N] = A, split extension");
        return [SplitExtension(pc, M)];
    fi;

    ## get all pairs of indices of generators of relations in <pc>
    ## and those positions that correspond to N x N except diagonal elements
    rels := AsSet(List(UnorderedTuples([1..Length(Pcgs(pc))], 2), Reversed));
    inds := List(pairs, pair-> Position(rels, pair));

    ## test whether generators of the group of cocycles allow any
    ## derived series extension; 
    ## the parts corresponding to N x N must at least generate A/[A,N]
    parts := List(TwoCocycles(pc, M),
                  c-> List(inds, ix -> c{[(ix-1)*d+1..ix*d]}));
    vs := VectorSpace(GF(p), Concatenation(Unique(Concatenation(parts)),
                                           AsList(subbasis)),
                      # provide the zero in case both lists are empty 
                      Zero(vs));
    if Dimension(vs) < d then
        Info(InfoSpecialExt, 1, "cocycles don't generate A/[A,N]");
        return EmptyPlist(0);
    fi;

    ## compute 2-cohomology
    cc := TwoCohomology(pc, M);

    ## special case: trivial 2-cohomology group 
    ## => only split extension can exist but this is a derived series
    ## extension if and only if A = [N, A], which has been dealt with already
    if Dimension(Image(cc.cohom)) = 0 then
        Info(InfoSpecialExt, 1, "trivial cohomology group");
        return EmptyPlist(0);
    fi;

    ## try whether the matrix representation of the group of compatible pairs
    ## is the same as that of the full direct product Aut(G) x Aut(A) and 
    ## can be obtained using pseudo random elements ...
    D := DirectProduct(AutomorphismGroup(pc), GL(d, p));
    mats := MatrixOperationOfCPGroup( cc, GeneratorsOfGroup( D ) );
    if mats <> fail then
        Mgrp := GroupByGenerators( mats );
        mats := [ ];
        for i in [1..10] do
            tup := PseudoRandom(D);
            if IsCompatiblePair(pc, M, tup) then
                Add(mats, MatrixOperationOfCP(cc, tup));
            fi;
            mats := Unique(mats);
            if Order(Mgrp) = Order(GroupByGenerators(mats)) then
                Mgrp := GroupByGenerators( mats );
                Info(InfoSpecialExt, 2, "Success, pseudo-random", Length(mats));
                break;
            fi;
        od;
    fi;

    ## ... otherwise
    if mats = fail or Order(Mgrp) > Order(GroupByGenerators(mats)) then
        ## calculate group of compatible pairs in matrix representation
        C := CompatiblePairs(pc, M);
        mats := MatrixOperationOfCPGroup( cc, GeneratorsOfGroup( C ) );
        Mgrp := GroupByGenerators( mats );
    fi;

    ## compute representatives of orbits of cocycles under the group of
    ## compatible pairs
    orbs := OrbitsDomain( Mgrp, Image(cc.cohom), OnRight );
    ccvecs := List( orbs, x -> PreImagesRepresentative( cc.cohom, x[1] ) );
 
    ## filter parts of the cocycles in <ccvecs> that correspond to N x N
    ## except diagonal entries corresponding to power relations
    result := List(ccvecs, c-> List(inds, ix -> c{[(ix-1)*d+1..ix*d]}));

    ## filter results that generate the whole of the module A together
    ## with the commutator subspace (thus generate A/[A,N])
    positions := EmptyPlist(d);
    for i in [1..Length(result)] do
        vs := VectorSpace(GF(p), Concatenation(result[i], AsList(subbasis)),
                          # provide the zero in case both lists are empty 
                          Zero(vs));
        if Dimension(vs) = d then
            Add(positions, i);
        fi;
    od;

    ## return derived series extensions
    return [ List(positions, i-> ExtensionSQ(cc.collector, pc, M, ccvecs[i])),
             true ]; 
end);
