
#
# computes up to isomorphism all lower central extensions (l.c.e.) of degree d
# of a nilpotent p-group G from the small groups library
#
InstallGlobalFunction(LowerCentralExtensionRepresentatives,
function( G, d )
    local low, # lower central series of <G>
          lambda1, # generators of first factor of <low>
          lambdac, # generators of last term of <low>
          p, # prime such that <G> is a p-group
          m, # rank <d> identity matrices, one for each generator of <G>
          M, # trivial degree <d> G-Module
          cc, # second cohomology record of <G> and <M>
          C, # compatible pairs of <G> and <M>
          mats, # matrices describing the operation of <C>
          Mgrp, # <C> as matrix group
          orbs, # orbits of H^2(G,M) under <Mgrp>
          reps, # one cocycle from each orbit in <orbs>
          result, # for each cocycle in <reps> a list containing module elements
          rels, # 2-tuples of generators of <G> (one for each relation)
          pair, # a tuple of one element in <lambda1> and one in <lambdac>
          pairs, # all <pair>s
          vecpos, # positions of relevant part of cocycles
          sigpos, # positions in <vecpos> where some cocycle is non-zero
          colmats, # list of list of parts corresponding to cocycle basis
          sigmats, # significant parts of cocycle basis transformed into mats
          basisZ2, # basis of the space of 2-cocycles <Source(cc.cohom)>
          sigparts, # parts of <basisZ2> at significant positions <sigpos>
          maxsubgrps, # maximal subgroups of G-module as vector space
          sum, # size of union of subspaces computed by IEP
          comb, # list indices of <maxsubgrps>
          schnitt, # intersection of <maxsubgrps{<comb>}>
          f, # homomorphism into factor space by <schnitt>
          proj, # projection of <sigparts> under <f> 
         # sub, # vector space spanned by <sigmats>
         # n, # maximum of number of rows or columns of a matrix in <sigmats>
         # r, # lower bound for the maximal rank of a matrix in <sub>
          pos, # position of <pair> in <rels>          
          i, # loop counter over cocycles in <reps>
          c, # i-th cocycle in <reps>
          positions, # positions of cocycles in <reps> that lead to l.c.e.
          ext, # list of extensions
          l, # number of maxmial subgroups of <M>
          dims, # list of list of counters of dimensions of intersections
          k, # loop counter, number of subgroups in intersection
          nr, # number of (k-1)-intersections of dimension <dim>
          dim, # loop counter, dimension
          total; # list of total counts of dimension of intersection

    ### MAIN FUNCTION BODY ##################################################

    # test group
    if not ( IsNilpotent(G) and IsPGroup(G) ) then
        Error("<G> has to be a nilpotent p-group");
    fi;

    ### MISSING: catch trivial case 
    if IsCyclic(G) then
        Print("cyclic group\n");
        return [ ];
    fi;

    # get the lower central series
    low := LowerCentralSeries(G);

    # test whether the pcgs runs through the series and adjust if necessary
    if not ForAll(Pcgs(low[Length(low)-1]), elm-> elm in Pcgs(G)) then
        Print("redefine \n");
        low := DerivedSeries(PcGroupToPcpGroup(G));
        G := PcpGroupToPcGroup(PcpGroupBySeries(low));
        low := LowerCentralSeries(G);        
    fi;

    # generators of the last term and of the first factor of the series
    lambdac := List(Pcgs(low[Length(low)-1]), x-> Position(Pcgs(G), x));
    lambda1 := List(Difference(Pcgs(low[1]), Pcgs(low[2])), 
                    x-> Position(Pcgs(G), x));

    ## test if the image can be big enough to have the full module dimension
    # AD improve test for abelian input groups
    if Length(lambdac) * Length(lambda1) < d then
        Print("dimension too large, \"h < d(A)\"\n");
        return [ ];
    fi;

    # pairs of numbers as identifiers for relations of G
    rels := AsSet(List(UnorderedTuples([1..Length(Pcgs(G))], 2), Reversed));

    # identifiers for relevant relations
    pairs := Unique(Filtered(List(Cartesian(lambdac, lambda1), 
                                  pp-> Reversed(Set(pp))),
                             set-> Length(set) = 2));

    # get relevant positions in cocycles
    vecpos := List(pairs, pair-> Position(rels, pair));

    # get parameters and cohomology information
    p := PrimePGroup( G );
    m := List( Pcgs(G), x-> IdentityMat( d, GF(p) ));
    M := GModuleByMats( m, GF(p) );
    cc := TwoCohomology( G, M );

    ### test whether lower central extensions are at all possible

    # collect relevant positions that are linearly independent
    colmats := List(vecpos, pos-> List(Basis(Image(cc.cohom)),
         x-> PreImagesRepresentative(cc.cohom,x){[(pos-1)*d+1..pos*d]}));

    sigpos := vecpos{LinearIndependentColumns(TransposedMat(List(colmats,
                                                        Concatenation)))};
    
    # the number of positions is smaller than the dimension of A
    if Length(sigpos) < d then
        Print("Basis exclusion 1\n");
        return [];
    fi;

    # collect the significant parts of the basis vectors as matrices ...
    sigmats := List(Basis(Source(cc.cohom)),
                    b-> List(sigpos, pos-> b{[(pos-1)*d+1..pos*d]}));;
    # ... remove duplicates and filter the zero matrices 
    sigmats := Filtered(Unique(sigmats), mat-> mat <> Zero(mat));;

    # the subgroup generated by all elements does not equal A
    if Rank(Concatenation(sigmats)) < d then
        Print("Basis exclusion 2\n");
        return [];
    fi;

    # compute the number of all cocycles with full relevant part
    basisZ2 := Basis(Source(cc.cohom));
    sigparts := List(basisZ2, c-> List(sigpos, pos-> c{[(pos-1)*d+1..pos*d]}));
    maxsubgrps := AsList(Subspaces(GF(p)^d, d-1));
    l := Length(maxsubgrps);

    dims := List([1..l], i-> ListWithIdenticalEntries(d,0)); 
    dims[1][d] := l;

    for k in [2..l] do
        # loop over dimension, actual dimension is <dim>-1 
        for dim in [1..d] do
            nr := dims[k-1][dim];
            if nr <> 0 then
                dims[k][dim] := dims[k][dim]
                                + ((p^(d+1-dim)-1)/(p-1) - (k-1)) * nr/k;
                if dim > 1 then
                    dims[k][dim-1] := dims[k][dim-1]
                                      + ((p^d-p^(d+1-dim))/(p-1)) * nr/k;
                fi;
            fi;
        od;
    od;

    # this is the d-th row in the inverse matrix of binomial coefficients
    total := Sum([1..l], i-> (-1)^(i+1)*dims[i]);
    sum := 0;
    for dim in [1..d] do
        proj := List(sigparts, l-> Concatenation(List(l, c-> c{[1..dim]})));
        sum := sum + total[d-dim+1] * p^Length(NullspaceMat(proj));
    od;

    # apply inclusion-exclusion-principle
    ## might be useful for the case that A is not elementary abelian
 #   sum := 0;
 #   for comb in Combinations([1..Length(maxsubgrps)]) do
 #       if Length(comb) <> 0 then
 #           schnitt := Intersection(maxsubgrps{comb});
 #           f := NaturalHomomorphismBySubspace(GF(p)^d, schnitt);
 #           proj := List(sigparts, l-> Concatenation(List(l, c-> c^f)));
 #           sum := sum + (-1)^(Length(comb)+1) * p^Length(NullspaceMat(proj));
 #       fi;
 #   od;

    if sum = p^Length(basisZ2) then
        Print("IEP exclusion\n");
        return [];
    fi;

#ausschnitt:=SortedList(Concatenation(List(vecpos,pos->[(pos-1)*d+1..pos*d])));
 
    # group of compatible pairs, full direct product as operation is trivial
    C := DirectProduct( AutomorphismGroup( G ), GL( d, p ) );
    mats := MatrixOperationOfCPGroup( cc, GeneratorsOfGroup( C ) );
    Mgrp := GroupByGenerators( mats );

    # compute orbit of mats on H^2( G, M )
    orbs := OrbitsDomain( Mgrp, Image(cc.cohom), OnRight );;
    reps := List( orbs, x -> PreImagesRepresentative( cc.cohom, x[1] ) );;

    # filter representatives that lead to lower central extensions
    result := List([1..Length(reps)], i-> [ ]);;

    for pair in pairs do
        pos := Position(rels, pair);

        for i in [1..Length(reps)] do
            c := reps[i];
            Add(result[i], c{[(pos-1)*d+1..pos*d]});
        od;
    od;

    # determine positions of cocycles and construct extensions
    positions := Filtered([1..Length(result)], i-> Rank(result[i]) = d);

    ext  := List( List(positions, i-> reps[i]),
                  x -> ExtensionSQ( cc.collector, G, M, x ) );

    return ext;
end);

#############################################################################
# 
# for a nilpotent p-group <G> all lower central extensions of elementary
# abelian p-groups by <G>
#

InstallGlobalFunction(AllLowerCentralExtensionRepresentatives,
function(G)
    local d,       # loop counter
          all,     # group extensions up to isomorphism 
          low,     # lower central series of <G>
          lambda1, # generators of first factor of <low>
          lambdac; # generators of last factor of <low>

    ### MAIN FUNCTION BODY ##################################################

    # test group
    if not ( IsNilpotent(G) and IsPGroup(G) ) then
        Error("<G> has to be a nilpotent p-group");
    fi;

    Info(InfoSpecialExt, 1, "Extensions of elementary abelian modules only.");
    all := [ ];

    # lower central series and its first and last term generators
    low := LowerCentralSeries(G);
    lambdac := List(Pcgs(low[Length(low)-1]), x-> Position(Pcgs(G), x));
    lambda1 := List(Difference(Pcgs(low[1]), Pcgs(low[2])), 
                    x-> Position(Pcgs(G), x));

    # the degree of lower central extensions is bounded
    Info(InfoSpecialExt, 2, "d_test: ",Length(lambdac) * Length(lambda1));
    for d in [ 1..Length(lambdac) * Length(lambda1) ] do
        Add( all, LowerCentralExtensionRepresentatives(G, d));
    od;

    return all;

end);

